#include "../global.h"

#include <time.h>

#include <Config.h>
#include <logger/Log.h>

#include <db/DoubleDatabase.h>
#include <itemstructs/Uint16ItemSet.h>

#include "../FicMain.h"
#include "../translate/TranslationRule.h"
#include "../translate/TranslationTable.h"
#include "ruletree/Pair.h"
#include "ruletree/RuleTree.h"

#include "ExactSearch.h"

ExactSearch::ExactSearch(Config *config) {
	mMaxPasses = config->Read<uint64>("maxPasses", 0);
	if(mMaxPasses==0) mMaxPasses = UINT64_MAX_VALUE;
	mMinSup = config->Read<uint32>("minSup", 2);
	if (mMinSup < 2) { mMinSup = 2; }
}
ExactSearch::~ExactSearch() {
	delete mVerticalDb;
	for(uint32 i=0; i<mNumPairs; i++)
		delete mPairs[i];
	delete[] mPairs;
	delete[] mItem2Alph;
	delete[] mAlph2Item;
	delete[] mNameDB;
}
void ExactSearch::Init(DoubleDatabase *ddb) {
	printf_s("> Initing search...\r");
	RemodelSearch::Init(ddb);

	uint16 alphSplit = ddb->GetAlphabetSplit();
	uint16 alphSize = ddb->GetAlphabetSize();

	const string nameDBTmp = ddb->GetLeftDatabase()->GetName(); 
	mNameDB = new char[nameDBTmp.size()+2];
	size_t length=nameDBTmp.copy(mNameDB, nameDBTmp.size()-3,0);
	mNameDB[length]='\0';


	// Construct vertical db, ie a tidset per singleton
	mVerticalDb = ddb->Transpose();
	mItems = (Uint16ItemSet**)(mVerticalDb->GetRows());	// each entry corresponds to an uint16 itemset -> the tidset
	alphabet *alph = ddb->GetAlphabet();
	alphabet::iterator it=alph->begin();
	double *stdLensL = ddb->GetLeftDatabase()->GetStdLengths();
	double *stdLensR = ddb->GetRightDatabase()->GetStdLengths();
	for(uint32 i=0; i<alphSize; i++) {  // foreach item
												// set the alphabet item of this particular item (important: array get shuffled later on, do not use index for this!)
		mItems[i]->SetID(i);
												// set its support (note: this is actually equivalent to the length of the tidset)
		mItems[i]->SetSupport(it->second);
												// set its singleton standard code length
		mItems[i]->SetStandardLength(i<alphSplit ? stdLensL[i] : stdLensR[i]);
		it++;
	}
	mAlph2Item = new uint16[alphSize];
	mItem2Alph = new uint16[alphSize];
	for(uint16 i=0; i<alphSize; i++)
		mAlph2Item[i] = mItem2Alph[i] = i;

	// Compute l<>r item pairs, their tidsets and their supports
	list<Pair*>	pairList;
	for(uint16 l=0; l<alphSplit; l++) {
		for(uint16 r=alphSplit; r<alphSize; r++) {
			if(mItems[l]->Intersects(mItems[r])) {
				Uint16ItemSet *tidSet = (Uint16ItemSet*)(mItems[l]->Intersection(mItems[r]));
				if(tidSet->GetLength() >= mMinSup)
					pairList.push_back(new Pair(l, r, tidSet));
				else
					delete tidSet;
			}
		}
	}
	mNumPairs = pairList.size();
	mPairs = new Pair *[mNumPairs];
	uint32 idx = 0;
	for(list<Pair*>::iterator it = pairList.begin(); it!=pairList.end(); it++)
		mPairs[idx++] = (Pair *)(*it);

	printf_s("> Found %u pairs that occur often enough (> %ux) in the data.\n", mNumPairs, mMinSup);
}

TranslationTable* ExactSearch::MineThatTable() {
	// Init file log
	OpenProgressFile();

	// Init TT
	TranslationTable *tt = new TranslationTable(mDDB);
	uint32 numRows = mDDB->GetLeftDatabase()->GetNumRows();
	uint32 numIL = mDDB->GetAlphabetSplit();	
	uint32 numIR = mDDB->GetRightDatabase()->GetAlphabetSize();
	tt->InitTranslation();

	// The Loop
	time_t searchStart = time(NULL);
	float gain;
	do {
		gain = 0.0;
		int startMs = getMilliCount();
		// Init search
		RuleTree *tree = new RuleTree(tt, mItems, mAlph2Item, mItem2Alph, mPairs, mNumPairs, mDDB->GetDataType());
		tree->SetMinSup(mMinSup);

		// Start search
		TranslationRule *rule = tree->MineBestRule(gain);

		// Add best rule found
		if(rule) {
			tt->Add(rule);
			printf_s(" # rules mined so far: %u\r", tt->GetNumRules());
			ProgressToFileAddRule(rule, 0, tt->GetCurStats(), getMilliSpan(startMs));
		}

		// Cleanup
		delete tree;
	} while(gain > 0.0 && tt->GetNumRules() < mMaxPasses);
	printf_s("> Stopped. # rules mined: %u\n", tt->GetNumRules());

	// Cleanup and roundup
	CloseProgressFile();

	char temp[500]; string s;
	time_t seconds = time(NULL) - searchStart;
#if defined (_MSC_VER) && defined (_WINDOWS)
	sprintf_s(temp, 500, "** Searching %s-cls-%ud with Exact dataSplit=%u, dataSize=%ux%u, numCands=%I64d(-0), searchTime=%d s, maxPasses=%I64d", mNameDB, mMinSup, numIL, numIR, numRows, mNumPairs, (uint32)seconds, mMaxPasses); s = temp; LOG_MSG(s);
#elif defined (__GNUC__) && defined (__unix__)
	sprintf_s(temp, 500, "** Searching %s-cls-%ud with Exact dataSplit=%u, dataSize=%ux%u, numCands=%lu(-0), searchTime=%d s, maxPasses=%lu", mNameDB, mMinSup, numIL, numIR, numRows, mNumPairs,seconds, mMaxPasses); s = temp; LOG_MSG(s);
#endif
	return tt;
}
