#include "../../global.h"

#include <Config.h>
#include <logger/Log.h>

#include <itemstructs/Uint16ItemSet.h>

#include "../../translate/TranslationRule.h"
#include "../../translate/TranslationTable.h"
#include "Pair.h"
#include "RuleTree.h"

#include "RuleNode.h"

RuleNode::RuleNode(RuleTree *tree, const uint32 startPairIdx) : mTree(tree) {
	uint32 alphSize = tree->mTT->GetAlphabetSize();
	Pair *startPair = tree->mPairs[startPairIdx];
	uint16 itemL = tree->mAlph2Item[startPair->mLeft];
	uint16 itemR = tree->mAlph2Item[startPair->mRight];

	mLeftSet = ItemSet::CreateSingleton(tree->mDataType, startPair->mLeft, alphSize);
	mRightSet = ItemSet::CreateSingleton(tree->mDataType, startPair->mRight, alphSize);

	mLeftTidSet = tree->mItems[itemL];
	mRightTidSet = tree->mItems[itemR];

	mExpandFromLeft = itemL + 1;
	mExpandFromRight = itemR + 1;
	
	mLeftStdSize = (float)(tree->mTT->mLeftStdLens[startPair->mLeft]);
	mRightStdSize = (float)(tree->mTT->mRightStdLens[startPair->mRight]);

	mToRightPotential = tree->mItemPotentials[itemL];
	mToLeftPotential = tree->mItemPotentials[itemR];
}
RuleNode::RuleNode(RuleTree *tree, const bool addToLeft, const uint16 addItem, const uint16 resumeItem, RuleNode *parent, Uint16ItemSet *leftTidSet, Uint16ItemSet *rightTidSet, const float changedPotential) : mTree(tree) {
	uint32 alphSize = tree->mTT->GetAlphabetSize();

	mLeftSet = parent->mLeftSet->Clone();
	mRightSet = parent->mRightSet->Clone();
	if(addToLeft) {
		mLeftSet->AddItemToSet(mTree->mItem2Alph[addItem]);
		mExpandFromLeft = addItem + 1;
		mExpandFromRight = resumeItem;
		mLeftStdSize = parent->mLeftStdSize + (float)(tree->mTT->mLeftStdLens[mTree->mItem2Alph[addItem]]);
		mRightStdSize = parent->mRightStdSize;

		mToLeftPotential = parent->mToLeftPotential;
		mToRightPotential = changedPotential;
	} else {
		mRightSet->AddItemToSet(mTree->mItem2Alph[addItem]);
		mExpandFromRight = addItem + 1;
		mExpandFromLeft = resumeItem;
		mLeftStdSize = parent->mLeftStdSize;
		mRightStdSize = parent->mRightStdSize + (float)(tree->mTT->mRightStdLens[mTree->mItem2Alph[addItem]]);

		mToLeftPotential = changedPotential;
		mToRightPotential = parent->mToRightPotential;
	}

	mLeftTidSet = leftTidSet;
	mRightTidSet = rightTidSet;
}
RuleNode::~RuleNode() {
	delete mLeftSet;
	delete mRightSet;
}
void RuleNode::MineBestRule(TranslationRule **bestRule, float &bestGain, const bool expandLeft /*= true*/) {
#if 0
	printf_s("- Considering %s <-> %s\n", mLeftSet->ToString(false, false).c_str(), mRightSet->ToString(false, false).c_str());
#endif
	double ruleSize = mLeftStdSize + mRightStdSize + 1.0;
	{
		// Start with a quick test, for this rule only
		float roughPotentialLtoR = (mLeftTidSet->GetLength()-1) * mRightStdSize - 2.0f;
		float roughPotentialRtoL = (mRightTidSet->GetLength()-1) * mLeftStdSize - 2.0f;
		float roughPotential = roughPotentialRtoL + roughPotentialLtoR + 3.0f;

		if(roughPotential > bestGain || roughPotentialRtoL > bestGain || roughPotentialLtoR > bestGain) { // Test for potential
			// Test <-,->,<-> rules with current left/right sets
			double gainToLeft, gainToRight;
			mTree->mTT->ComputeGainForAdditionFromTree(mLeftSet, mRightSet, mLeftTidSet, mRightTidSet, gainToRight, gainToLeft);

			// Subtract rule sizes
			double gainToBoth = gainToLeft + gainToRight - ruleSize;
			gainToLeft -= (ruleSize + 1.0); // extra bit penalty for direction
			gainToRight -= (ruleSize + 1.0); // extra bit penalty for direction

			if(gainToBoth > bestGain && gainToBoth > gainToLeft && gainToBoth > gainToRight) {
				delete (*bestRule);
				(*bestRule) = new TranslationRule(mLeftSet->Clone(), RuleBoth, mRightSet->Clone(), ruleSize);
				bestGain = (float)gainToBoth;
				printf_s("\t! %s (%.02fb)\n", (*bestRule)->ToString().c_str(), bestGain);
			} else if(gainToLeft > bestGain && gainToLeft > gainToRight) {
				delete (*bestRule);
				(*bestRule) = new TranslationRule(mLeftSet->Clone(), RuleToLeft, mRightSet->Clone(), ruleSize + 1.0);
				bestGain = (float)gainToLeft;
				printf_s("\t! %s (%.02fb)\n", (*bestRule)->ToString().c_str(), bestGain);
			} else if(gainToRight > bestGain) {
				delete (*bestRule);
				(*bestRule) = new TranslationRule(mLeftSet->Clone(), RuleToRight, mRightSet->Clone(), ruleSize + 1.0);
				bestGain = (float)gainToRight;
				printf_s("\t! %s (%.02fb)\n", (*bestRule)->ToString().c_str(), bestGain);
			}
		}
	}

	if(expandLeft) {
		// Expand left side in all possible ways
		uint32 upToItem = mTree->mTT->GetAlphabetSplit();
		for(uint32 i=mExpandFromLeft; i<upToItem; i++) {
			// support checks
			if(mTree->mItems[i]->GetLength() < mTree->mMinSup)
				continue;
			Uint16ItemSet *leftTidSet = (Uint16ItemSet *)(mLeftTidSet->Intersection(mTree->mItems[i]));
			if(leftTidSet->GetLength() < mTree->mMinSup) { // new left side has too low support, do not expand
				delete leftTidSet;
				continue;
			}

			// Potential gain check; only need to compute in the direction that has been changed
			float *potentials = mTree->mRowPotentials[RIGHT];
			float toRightPotential = 0.0f;
#define POTENTIAL_CHECK	// Comment this line if you want to disable potential gain checking -- may be faster in some cases, in particular with high minSups
							// (Actually, it might be useful to make this a config setting?)
#ifdef POTENTIAL_CHECK
			uint32 len = leftTidSet->GetLength();
			uint16 *rows = leftTidSet->GetSet();
			for(uint32 r=0; r<len; r++)
				toRightPotential += potentials[rows[r]];
			if(mToLeftPotential + toRightPotential - ruleSize < bestGain) { // two-way potential lower than current best gain
				delete leftTidSet;
				continue;
			}
#endif

			// Expand
			RuleNode *node = new RuleNode(mTree, true, i, mExpandFromRight, this, leftTidSet, mRightTidSet, toRightPotential);
			node->MineBestRule(bestRule, bestGain);
			delete node;

			// Cleanup tidset
			delete leftTidSet;
		}
	}

	// Expand right side in all possible ways
	uint32 upToItem = mTree->mTT->GetAlphabetSize();
	for(uint32 i=mExpandFromRight; i<upToItem; i++) {
		// support check
		if(mTree->mItems[i]->GetLength() < mTree->mMinSup)
			continue;
		Uint16ItemSet *rightTidSet = (Uint16ItemSet *)(mRightTidSet->Intersection(mTree->mItems[i]));
		if(rightTidSet->GetLength() < mTree->mMinSup) { // new right side has too low support, do not expand
			delete rightTidSet;
			continue;
		}

		// Potential gain check; only need to compute in the direction that has been changed
		float *potentials = mTree->mRowPotentials[LEFT];
		float toLeftPotential = 0.0f;
#ifdef POTENTIAL_CHECK
		uint32 len = rightTidSet->GetLength();
		uint16 *rows = rightTidSet->GetSet();
		for(uint32 r=0; r<len; r++)
			toLeftPotential += potentials[rows[r]];
		if(toLeftPotential + mToRightPotential - ruleSize < bestGain) { // two-way potential lower than current best gain
			delete rightTidSet;
			continue;
		}
#endif
		// Expand
		RuleNode *node = new RuleNode(mTree, false, i, mExpandFromLeft, this, mLeftTidSet, rightTidSet, toLeftPotential);
		node->MineBestRule(bestRule, bestGain, false); // do not expand left side from here on
		delete node;

		// Cleanup tidset
		delete rightTidSet;
	}
}
