#include "../../global.h"

#include <Config.h>
#include <logger/Log.h>

#include <itemstructs/Uint16ItemSet.h>

#include "../../translate/TranslationRule.h"
#include "../../translate/TranslationTable.h"
#include "Pair.h"
#include "RuleNode.h"

#include "RuleTree.h"

RuleTree::RuleTree(TranslationTable *tt, Uint16ItemSet **items, uint16 *alph2Item, uint16 *item2Alph, Pair **pairs, const uint32 numPairs, ItemSetType dataType) {
	// Init
	mMinSup = 2;
	mTT = tt;
	mItems = items;
	mAlph2Item = alph2Item;
	mItem2Alph = item2Alph;
	mPairs = pairs;
	mNumPairs = numPairs;
	mDataType = dataType;

	// Get gain potentials per row for current uncovered state
	mRowPotentials = tt->ComputeRowGainPotentials();

	// Compute expansion potentials for each pair
	for(uint32 i=0; i<mNumPairs; i++) {
		Pair *p = mPairs[i];

		// mItems[p->mLeft] contains tidSet of left item: l->r applies to those rows
		p->mLtoRPotential = -2.0f; // two bits for direction
		uint32 len = mItems[mAlph2Item[p->mLeft]]->GetLength();
		uint16 *rows = mItems[mAlph2Item[p->mLeft]]->GetSet();
		for(uint32 r=0; r<len; r++)
			p->mLtoRPotential += mRowPotentials[RIGHT][rows[r]];

		// mItems[p->mRight] contains tidSet of right item: l<-r applies to those rows
		p->mRtoLPotential = -2.0f; // two bits for direction
		len = mItems[mAlph2Item[p->mRight]]->GetLength();
		rows = mItems[mAlph2Item[p->mRight]]->GetSet();
		for(uint32 r=0; r<len; r++)
			p->mRtoLPotential += mRowPotentials[LEFT][rows[r]];

		p->mPotential = p->mLtoRPotential + p->mRtoLPotential + 3.0f; // just one bit penalty, for indicating `both directions'
	}

	// Bubble-sort pairs on expansion potentials descending
	printf_s("> Sorting pairs...\r");
	for(uint32 i=1; i<mNumPairs; i++) {
		for(uint32 j=0; j<mNumPairs-i; j++) {
			if(mPairs[j]->mPotential < mPairs[j+1]->mPotential) {
				Pair *tmp = mPairs[j];
				mPairs[j] = mPairs[j+1];
				mPairs[j+1] = tmp;
			}
		}
	}
	printf_s("> Sorting pairs... done\n");

	// Debug: output sorted pairs and their potentials, if you like
#if 0
	for(uint32 i=0; i<mNumPairs; i++)
		printf_s("(%u,%u) has potential %f\n", mPairs[i]->mLeft, mPairs[i]->mRight, mPairs[i]->mPotential);
#endif

	// Compute expansion potentials for each item and sort items accordingly
	{
		uint16 alphSize = tt->GetAlphabetSize();
		uint16 alphSplit = tt->GetAlphabetSplit();
		mItemPotentials = new float[alphSize];
		for(uint32 i=0; i<alphSplit; i++) { // l->r
			mItemPotentials[i] = 0.0f;
			uint32 len = mItems[i]->GetLength();
			uint16 *rows = mItems[i]->GetSet();
			for(uint32 r=0; r<len; r++)
				mItemPotentials[i] += mRowPotentials[RIGHT][rows[r]];
		}
		for(uint32 i=alphSplit; i<alphSize; i++) { // l<-r
			mItemPotentials[i] = 0.0f;
			uint32 len = mItems[i]->GetLength();
			uint16 *rows = mItems[i]->GetSet();
			for(uint32 r=0; r<len; r++)
				mItemPotentials[i] += mRowPotentials[LEFT][rows[r]];
		}

		// Bubble-sort items on expansion potentials descending
		printf_s("> Sorting items...\r");
		for(uint32 i=1; i<alphSplit; i++) { // left side
			for(uint32 j=0; j<alphSplit-i; j++) {
				if(mItemPotentials[j] < mItemPotentials[j+1]) {
					Uint16ItemSet *tmp = mItems[j];
					mItems[j] = mItems[j+1];
					mItems[j+1] = tmp;
					float tmp1 = mItemPotentials[j];
					mItemPotentials[j] = mItemPotentials[j+1];
					mItemPotentials[j+1] = tmp1;
				}
			}
		}
		uint32 rightAlph = alphSize - alphSplit;
		for(uint32 i=1; i<rightAlph; i++) { // right side
			for(uint32 j=0; j<rightAlph-i; j++) {
				if(mItemPotentials[alphSplit+j] < mItemPotentials[alphSplit+j+1]) {
					Uint16ItemSet *tmp = mItems[alphSplit+j];
					mItems[alphSplit+j] = mItems[alphSplit+j+1];
					mItems[alphSplit+j+1] = tmp;
					float tmp1 = mItemPotentials[alphSplit+j];
					mItemPotentials[alphSplit+j] = mItemPotentials[alphSplit+j+1];
					mItemPotentials[alphSplit+j+1] = tmp1;
				}
			}
		}
		printf_s("> Sorting items... done\n");

#if 0
		for(uint32 i=0; i<alphSize; i++)
			printf_s("%u has potential %f\n", (uint32)mItems[i]->GetID(), mItemPotentials[i]);
#endif

		for(uint32 i=0; i<alphSize; i++) {
			mItem2Alph[i] = (uint16) mItems[i]->GetID();
			mAlph2Item[mItems[i]->GetID()] = i;
		}
	}
	// IMPORTANT: cannot assume mItems to be sorted on alphabet (accessible as mItems[i]->GetID())
	// Use mItem2Alph and mAlph2Item maps instead
}
RuleTree::~RuleTree() {
	delete[] mItemPotentials;

	delete[] mRowPotentials[LEFT];
	delete[] mRowPotentials[RIGHT];
	delete[] mRowPotentials;
}
TranslationRule* RuleTree::MineBestRule(float &gain) {
	TranslationRule *rule = NULL;

	printf_s("> MineBestRule...\n");

	// Traverse rule tree
	for(uint32 i=0; i<mNumPairs; i++) {
	//for(uint32 i=0; i<1; i++) {
		RuleNode *node = new RuleNode(this, i);
		node->MineBestRule(&rule, gain);
		delete node;
	}
	printf_s("> MineBestRule... done\n");
	return rule;
}
